/*
 * Copyright (c) 2008, Desiderata Software
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 * Redistributions of source code must retain the above copyright notice, this list
 * of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice, this
 * list of conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution. 
 * 
 * Neither the name of Desiderata Software nor the names of its contributors may be
 * used to endorse or promote products derived from this software without specific
 * prior written permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *
 * Module Name:           mod_jaxcent.cpp
 * Original Author:       Mukesh Prasad
 * Brief Description:     Apache Connector.
 * Change History:
 *
 *     1/23/2008  MP      Initial version.
 *     7/12/2008  MP      Modified for file uploads, virtual-host support.
 *
 */

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "http_request.h"
#include "util_script.h"
#include "http_connection.h"

#include "apr_strings.h"

#define APR_WANT_STRFUNC
#define APR_WANT_BYTEFUNC
#define APR_WANT_STDIO
#define APR_WANT_MEMFUNC

#include "apr_want.h"

#include <stdarg.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "jni.h"
#include "jni_md.h"

#ifndef _WIN32
#include <unistd.h>
#include <dlfcn.h>
#endif

// Struct to hold virtual-host specific parameters.

struct ParamList {
    server_rec * server;
    char * key;
    char * value;
    ParamList * link;
};

// Struct to hold list of virtual servers

struct ServerList {
    server_rec * server;
    char * error;
    jobject connector;
    ServerList * next;
};

// The JVM, classpath, and log file are shared for the entire connectors,
// and cannot be switched between virtual hosts. (These are not virtual-host specific.)

static char * jvm = NULL;              // Path to JVM
static char * classpath = NULL;        // Classpath for the JVM
static char * logfile = NULL;          // Output log file

static ParamList * params = NULL;      // List of virtual-host specific parameters
static ParamList * lastParam = NULL;   // Last parameter, for list manipulation
static ServerList * serverList = NULL; // List of virtual hosts

static bool inited = false;            // Has Java been initialized already?
static char * error = NULL;            // Last error encountered

typedef jint (JNICALL * CreateJVM)(JavaVM **, void **, JavaVMInitArgs *);
                                // Loaded method for creating JVM

typedef const char *(* CmdFn)(void);
static JavaVM * vm = 0;                // Pointer to JVM instance

// Method IDs for various Java methods
static jmethodID processMethod;
static jmethodID uploadInitMethod;
static jmethodID uploadMethod;
static jmethodID uploadEndMethod;

static jmethodID terminateMethod;

// Global ref to java.lang.Object
static jclass jobjectClass;

// Cookie for Jaxcent sessions
static const char jaxcentCookieId[] = "JAXCENT_APACHE_SESSIONID";

// Forward declarations.

static const char * set_jvm(cmd_parms *cmd, void *dummy, const char *arg);
static const char * set_classpath(cmd_parms *cmd, void *dummy, const char *arg);
static const char * set_config_file(cmd_parms *cmd, void *dummy, const char *arg);
static const char * set_log_file(cmd_parms *cmd, void *dummy, const char *arg);
static const char * set_reloadable_classes(cmd_parms *cmd, void *dummy, const char *arg);
static const char * set_display_servlet_message(cmd_parms *cmd, void *dummy, const char *arg);
static const char * set_refresh_url_map(cmd_parms *cmd, void *dummy, const char *arg);
static const char * set_keep_alive_timeout(cmd_parms *cmd, void *dummy, const char *arg);
static const char * set_parameter(cmd_parms *cmd, void *dummy, const char *arg);
static void terminate_module();

static void register_hooks(apr_pool_t *p);

// The parameter definitions.

static const command_rec jaxcent_cmds[] =
{
    AP_INIT_RAW_ARGS("JVM", (CmdFn) set_jvm, NULL, RSRC_CONF|ACCESS_CONF, "Path to JVM"),
    AP_INIT_RAW_ARGS("classpath", (CmdFn) set_classpath, NULL, RSRC_CONF|ACCESS_CONF, "Classpath for JVM"),
    AP_INIT_RAW_ARGS("LogFile", (CmdFn) set_log_file, NULL, RSRC_CONF|ACCESS_CONF, "Logfile for error messages"),
    AP_INIT_RAW_ARGS("JaxcentConfigXML", (CmdFn) set_config_file, NULL, RSRC_CONF|ACCESS_CONF, "Jaxcent configuration XML file"),
    AP_INIT_RAW_ARGS("ReloadableClasses", (CmdFn) set_reloadable_classes, NULL, RSRC_CONF|ACCESS_CONF, "Jaxcent Reloadable classpath"),
    AP_INIT_RAW_ARGS("DisplayServletMessage", (CmdFn) set_display_servlet_message, NULL, RSRC_CONF|ACCESS_CONF, "Whether to display standard message or error 404"),
    AP_INIT_RAW_ARGS("RefreshURLMap", (CmdFn) set_refresh_url_map, NULL, RSRC_CONF|ACCESS_CONF, "Whether to re-read the config XML file when it changes"),
    AP_INIT_RAW_ARGS("KeepAliveTimeout", (CmdFn) set_keep_alive_timeout, NULL, RSRC_CONF|ACCESS_CONF, "Keep alive timeout in seconds"),
    AP_INIT_RAW_ARGS("Parameter", (CmdFn) set_parameter, NULL, RSRC_CONF|ACCESS_CONF, "Parameter(s) to be passed to the Jaxcent framework"),
    
    { NULL }
};

// Standard apache module definition.
extern "C" module AP_MODULE_DECLARE_DATA jaxcent_module = {
    STANDARD20_MODULE_STUFF,
    NULL,                       /* create per-directory config structure */
    NULL,                       /* merge per-directory config structures */
    NULL,                       /* create per-server config structure */
    NULL,                       /* merge per-server config structures */
    jaxcent_cmds,               /* command apr_table_t */
    register_hooks              /* register hooks */
};

// set_error sets an error message that will be displayed by the server.

static void set_error( char ** error, JNIEnv * jenv, const char * fmt, ... )
{
    va_list marker;
    char buffer[ 2000 ];

    va_start( marker, fmt );
    int len = vsprintf( buffer, fmt, marker );
    va_end( marker );
    if ( *error != NULL )
        delete[] *error;
    if ( jenv ) {
        jthrowable ex = jenv->ExceptionOccurred();
        if ( ex ) {
            jenv->ExceptionClear();
            jclass bacls = jenv->FindClass( "java/io/ByteArrayOutputStream" );
            jmethodID bactor = jenv->GetMethodID( bacls, "<init>", "()V" );
            jmethodID tobytes = jenv->GetMethodID( bacls, "toByteArray", "()[B" );
            jobject ba = jenv->NewObject( bacls, bactor );
            jclass pscls = jenv->FindClass( "java/io/PrintStream" );
            jmethodID psctor = jenv->GetMethodID( pscls, "<init>", "(Ljava/io/OutputStream;)V" );
            jmethodID psclose = jenv->GetMethodID( pscls, "close", "()V" );
            jobject ps = jenv->NewObject( pscls, psctor, ba );
            jclass excls = jenv->GetObjectClass( ex );
            jmethodID printStackTrace = jenv->GetMethodID( excls, "printStackTrace", "(Ljava/io/PrintStream;)V" );
            jenv->CallVoidMethod( ex, printStackTrace, ps );
            jenv->CallVoidMethod( ps, psclose );
            jbyteArray jba = (jbyteArray) jenv->CallObjectMethod( ba, tobytes );
            jbyte * bytes = jenv->GetByteArrayElements( jba, 0 );
			int blen = jenv->GetArrayLength( jba );
            *error = new char[ strlen( buffer ) + 2 + blen + 1 ];
            strcpy( *error, buffer );
            strcat( *error, "\r\n" );
			char * err = *error;
			int len = strlen( err );
			memcpy( &err[len], bytes, blen );
			err[len+blen] = '\0';
            jenv->ReleaseByteArrayElements( jba, bytes, 0 );
            jenv->DeleteLocalRef( jba );
            jenv->DeleteLocalRef( ps );
            jenv->DeleteLocalRef( ba );
            jenv->DeleteLocalRef( ex );
            return;
        }
        
    }
    *error = new char[ strlen( buffer ) + 1 ];
    strcpy( *error, buffer );
}

// Find a given server (virtual host) or insert a new record.
static ServerList * getServer( server_rec * server )
{
    ServerList * s = serverList;
    while ( s != NULL ) {
        if ( s->server == server )
            break;
        s = s->next;
    }
    if ( s == NULL ) {
        s = new ServerList;
        s->server = server;
        s->connector = NULL;
        s->error = NULL;
        s->next = serverList;
        serverList = s;
    }
    return s;
}

// Add a parameter and value for a given virtual host.

static void add_parameter( server_rec * server, const char * key, const char * value )
{
    ParamList * pl = new ParamList;
    pl->server = server;
    pl->key = strcpy( new char[ strlen( key ) + 1 ], key );
    pl->value = strcpy( new char[ strlen( value ) + 1 ], value );
    pl->link = NULL;
    if ( lastParam != NULL )
        lastParam->link = pl;
    else
        params = pl;
    lastParam = pl;
    if ( strcasecmp( key, "jvm" ) == 0 )
        jvm = pl->value;
    else if ( strcasecmp( key, "classpath" ) == 0 )
        classpath = pl->value;
    else if ( strcasecmp( key, "logfile" ) == 0 )
        logfile = pl->value;
}

// Initialize Java
static void init_java_connection()
{
    if ( inited )
        return;
    inited = true;

    // Check all required parameters have been specified in httpd.conf
    // If not, set error message and return.

    if ( jvm == NULL ) {
        set_error( &error, NULL, "JVM has not been set in the configuration file" );
        return;
    }
#ifdef _WIN32
    if ( strlen( jvm ) < 5 ||
        strcasecmp( &jvm[ strlen(jvm)-4 ], ".dll" ) != 0 ) {
        set_error( &error, NULL, "JVM must be a .DLL file. (Found \"%s\")", jvm );
        return;
    }
#endif
    if ( classpath == NULL ) {
        set_error( &error, NULL, "CLASSPATH has not been specified in the configuration file" );
        return;
    }
    if ( logfile != NULL ) {
	    if ( access( logfile, 0 ) != 0 ) {
		    FILE * fp = fopen( logfile, "wb" );
			if ( fp )
				fclose( fp );
		}
		if ( access( logfile, 02 ) != 0 ) {
			set_error( &error, NULL, "Cannot write LogFile \"%s\"", logfile );
			return;
		}
    }
    // Everything ok.
    // Load the JVM.
#ifdef _WIN32
    HMODULE jvmLib = LoadLibrary( jvm );
    if ( jvmLib == NULL ) {
        void * lpMsgBuf;
        if ( FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
            NULL, GetLastError(), MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),(LPTSTR) &lpMsgBuf, 0, 0 ) == 0 )
            set_error( &error, NULL, "Could not load JVM \"%s\"", jvm );
        else {
            set_error( &error, NULL, "Could not load JVM \"%s\": %s", jvm, lpMsgBuf );
            LocalFree( lpMsgBuf );
        }
        return;
    }
    CreateJVM createJVM = (CreateJVM) ::GetProcAddress( jvmLib, "JNI_CreateJavaVM" );
#else
    void * jvmLib = dlopen( jvm, RTLD_LAZY );
    if ( jvmLib == NULL ) {
        set_error( &error, NULL, "Could not load JVM \"%s\"", jvm );
        return;
    }
    CreateJVM createJVM = (CreateJVM) dlsym( jvmLib, "JNI_CreateJavaVM" );
#endif
    if ( createJVM == NULL ) {
        set_error( &error, NULL, "JVM Library \"%s\" does not contain a valid JVM", jvm );
        return;
    }
    JavaVMInitArgs vm_args;
    vm_args.version = JNI_VERSION_1_2;
    JavaVMOption optClasspath;
    char * classpathOption = new char[ strlen( classpath ) + 22 ];
    strcpy( classpathOption, "-Djava.class.path=" );
    strcat( classpathOption, classpath );
    optClasspath.optionString = classpathOption;
    optClasspath.extraInfo = 0;
    vm_args.nOptions = 1;
    vm_args.options = &optClasspath;
    JNIEnv * jenv;
    int result = createJVM( &vm, (void **) &jenv, &vm_args );
    delete[] classpathOption;
    if ( result < 0 ) {
        set_error( &error, NULL, "Java VM Creation from \"%s\" failed", jvm );
        return;
    }
    // Java VM creation succeeded.

    // Now load the Jaxcent object
    if ( jenv->FindClass( "jaxcentFramework/JaxcentFramework" ) == NULL ) {
        set_error( &error, jenv, "Could not find class \"jaxcentFramework.JaxcentFramework\" in classpath\r\n     \"%s\"\r\nMake sure to include path to jaxcentFramework JAR file.", classpath );
        return;
    }
    jclass connectorClass = jenv->FindClass( "jaxcentConnector/JaxcentHttpConnector" );
    if ( connectorClass == NULL ) {
        set_error( &error, jenv, "Could not find class \"jaxcentConnector.JaxcentHttpConnector\" in classpath\r\n     \"%s\"\r\nMake sure to include path to jaxcentConnector JAR file.", classpath );
        return;
    }
    // Get the ctor.
    jmethodID ctor = jenv->GetMethodID( connectorClass, "<init>", "([Ljava/lang/String;[Ljava/lang/String;)V" );
    if ( ctor == NULL ) {
        set_error( &error, jenv, "Internal error in class jaxcentConnector.JaxcentHttpConnector (ctor not found)" );
        return;
    }
    // Locate other methods required from the connector class.
    jmethodID initMethod = jenv->GetMethodID( connectorClass, "init", "()V" );
    if ( initMethod == NULL ) {
        set_error( &error, jenv, "Internal error in class jaxcentConnector.JaxcentHttpConnector (\"init\" not found)" );
        return;
    }
    processMethod = jenv->GetMethodID( connectorClass, "process", "([Ljava/lang/Object;Ljava/lang/String;Ljava/lang/String;)V" );
    if ( processMethod == NULL ) {
        set_error( &error, jenv, "Internal error in class jaxcentConnector.JaxcentHttpConnector (\"init\" not found)" );
        return;
    }
    terminateMethod = jenv->GetMethodID( connectorClass, "terminate", "()V" );
    if ( terminateMethod == NULL ) {
        set_error( &error, jenv, "Internal error in class jaxcentConnector.JaxcentHttpConnector (\"terminate\" not found)" );
        return;
    }
    uploadInitMethod = jenv->GetMethodID( connectorClass, "onUploadInit", "(Ljava/lang/String;Ljava/lang/String;I)Ljava/lang/Object;" );
    if ( jenv->ExceptionOccurred()) {
        uploadInitMethod = 0;
        jenv->ExceptionClear();
    }
    uploadEndMethod = jenv->GetMethodID( connectorClass, "onUploadEnd", "(Ljava/lang/Object;)[B" );
    if ( jenv->ExceptionOccurred()) {
        uploadEndMethod = 0;
        jenv->ExceptionClear();
    }
    uploadMethod = jenv->GetMethodID( connectorClass, "onUpload", "(Ljava/lang/Object;[B)Z" );
    if ( jenv->ExceptionOccurred()) {
        uploadMethod = 0;
        jenv->ExceptionClear();
    }
    jmethodID logFileMethod = jenv->GetStaticMethodID( connectorClass, "setLogfile", "(Ljava/lang/String;)Ljava/lang/String;" );
    if ( logFileMethod == NULL ) {
        set_error( &error, jenv, "Internal error in class jaxcentConnector.JaxcentHttpConnector (\"setLogfile\" not found)" );
        return;
    }
	if ( logfile ) {
		// Initialize log file
		jstring jlogfile = jenv->NewStringUTF( logfile );
		jstring jlogresult = (jstring) jenv->CallStaticObjectMethod( connectorClass, logFileMethod, jlogfile );
		if ( jlogresult != NULL ) {
			const char * logresult = jenv->GetStringUTFChars( jlogresult, NULL );
			set_error( &error, jenv, logresult );
			jenv->ReleaseStringUTFChars( jlogresult, logresult );
			jenv->DeleteLocalRef( jlogresult );
		}
		jenv->DeleteLocalRef( jlogfile );
	}
    jobjectClass = (jclass) jenv->NewGlobalRef( jenv->FindClass( "java/lang/Object" ));
}

// Get the connector object for the given virtual host.

static jobject getConnector( JNIEnv * jenv, server_rec * server )
{
    ServerList * s = getServer( server );
    if ( s->connector )
        return s->connector;
    if ( s->error ) {
        return NULL;
    }

    // Instantiate a new connector.

    int nCount = 0;
    ParamList * pl;
    for ( pl = params; pl; pl = pl->link ) {
        if ( pl->server == NULL || pl->server == server ) {
            nCount++;
        }
    }
    jclass connectorClass = jenv->FindClass( "jaxcentConnector/JaxcentHttpConnector" );
    if ( connectorClass == NULL ) {
        set_error( &s->error, jenv, "Could not find class \"jaxcentConnector.JaxcentHttpConnector\" in classpath\r\n     \"%s\"\r\nMake sure to include path to jaxcentConnector JAR file.", classpath );
        return NULL;
    }
    jmethodID ctor = jenv->GetMethodID( connectorClass, "<init>", "([Ljava/lang/String;[Ljava/lang/String;)V" );
    if ( ctor == NULL ) {
        set_error( &s->error, jenv, "Internal error in class jaxcentConnector.JaxcentHttpConnector (ctor not found)" );
        return NULL;
    }
    jmethodID initMethod = jenv->GetMethodID( connectorClass, "init", "()V" );
    if ( initMethod == NULL ) {
        set_error( &s->error, jenv, "Internal error in class jaxcentConnector.JaxcentHttpConnector (\"init\" not found)" );
        return NULL;
    }
    jclass stringClass = jenv->FindClass( "java/lang/String" );
    jobjectArray keys = jenv->NewObjectArray( nCount, stringClass, NULL );
    jobjectArray values = jenv->NewObjectArray( nCount, stringClass, NULL );
    int index = 0;
    const char * configFile = NULL;
    if ( server != NULL ) {
        for ( pl = params; pl; pl = pl->link ) {
            if ( pl->server == server ) {
                if ( strcasecmp( pl->key, "JaxcentConfigXML" ) == 0 )
                    configFile = pl->value;
                jenv->SetObjectArrayElement( keys, index, jenv->NewStringUTF( pl->key ));
                jenv->SetObjectArrayElement( values, index, jenv->NewStringUTF( pl->value ));
                index++;
            }
        }
    }
    for ( pl = params; pl; pl = pl->link ) {
        if ( pl->server == NULL ) {
            if ( configFile == NULL && strcasecmp( pl->key, "JaxcentConfigXML" ) == 0 )
                configFile = pl->value;
            jenv->SetObjectArrayElement( keys, index, jenv->NewStringUTF( pl->key ));
            jenv->SetObjectArrayElement( values, index, jenv->NewStringUTF( pl->value ));
            index++;
        }
    }
    if ( configFile == NULL ) {
        set_error( &s->error, jenv, "JaxcentConfigXML has not been specified in the configuration file" );
        return NULL;
    }
    if ( access( configFile, 04 ) != 0 ) {
        set_error( &s->error, jenv, "Cannot read JaxcentConfigXML file \"%s\"", configFile );
        return NULL;
    }
    s->connector = jenv->NewObject( connectorClass, ctor, keys, values );
    if ( s->connector == NULL ) {
        set_error( &s->error, jenv, "Internal error: Could not instantiate jaxcentConnector.JaxcentHttpConnector" );
        return NULL;
    }
    s->connector = jenv->NewGlobalRef( s->connector );
    jenv->CallVoidMethod( s->connector, initMethod );
    return s->connector;
}

// Handle "Upload"s from posted data.

static int processUpload( JNIEnv * jenv, request_rec * r, const char * contentType, jobject connector )
{
    jstring jargs = 0;
    if ( r->args && *r->args ) {
        jargs = jenv->NewStringUTF( r->args );
    }
    const char * contentLengthStr = apr_table_get(r->headers_in, "Content-Length");
    int contentLength = -1;
    if ( contentLengthStr && *contentLengthStr )
        contentLength = atoi( contentLengthStr );

    ap_set_content_type(r, "text/plain");
    jstring jcontentType = jenv->NewStringUTF( contentType );
    jobject ref = jenv->CallObjectMethod( connector, uploadInitMethod, jargs, jcontentType, contentLength );
    jenv->DeleteLocalRef( jcontentType );
    apr_bucket_brigade * brigade = apr_brigade_create( r->pool, r->connection->bucket_alloc );

    bool send = true;
    bool seen_eos = false;

    do {
        apr_bucket *bucket;

        if ( ap_get_brigade(r->input_filters, brigade,
                                AP_MODE_READBYTES, APR_BLOCK_READ,
                                8192) != APR_SUCCESS )
                                break;

        for (bucket = APR_BRIGADE_FIRST(brigade);
             bucket != APR_BRIGADE_SENTINEL(brigade);
             bucket = APR_BUCKET_NEXT(bucket))
        {
            const char *data;
            apr_size_t len;

            if (APR_BUCKET_IS_EOS(bucket)) {
                seen_eos = true;
                break;
            }

            if (APR_BUCKET_IS_METADATA(bucket)) {
                continue;
            }

            if (  apr_bucket_read(bucket, &data, &len, APR_BLOCK_READ) != APR_SUCCESS )
                break;

            if ( len > 0 && data != 0 && send ) {
                jbyteArray jdata = jenv->NewByteArray( len );
                jbyte * jbuf = jenv->GetByteArrayElements( jdata, NULL );
                memcpy( jbuf, data, len );
                jenv->ReleaseByteArrayElements( jdata, jbuf, 0 );
                if ( jenv->CallBooleanMethod( connector, uploadMethod, ref, jdata ) == JNI_FALSE )
                    send = false;
                jenv->DeleteLocalRef( jdata );
            }
        }

        apr_brigade_cleanup(brigade);
    } while (!seen_eos);

    apr_brigade_destroy(brigade);

//    apr_brigade_cleanup(bb);

    jbyteArray resp = NULL;
    if ( uploadEndMethod != NULL ) {
        resp = (jbyteArray) jenv->CallObjectMethod( connector, uploadEndMethod, ref );
    }

    if ( resp != NULL ) {
        // Trying to send it back results in occasional "NOT IMPLEMENTED" errors.
        // Sending the response back is not important anyway.  So we just leave
        // the response as an empty string.

        /* int len = jenv->GetArrayLength( resp );
        if ( len > 0 ) {
            jbyte * postResult = jenv->GetByteArrayElements( resp, 0 );
            ap_rwrite( postResult, len, r );
            jenv->ReleaseByteArrayElements( resp, postResult, 0 );
        }
        */
        jenv->DeleteLocalRef( resp );
    }

    if ( jargs )
        jenv->DeleteLocalRef( jargs );

    if ( ref )
        jenv->DeleteLocalRef( ref );
    return OK;
}

// Main handler entry point.

static int jaxcent_handler(request_rec *r)
{
    if (strcmp(r->handler, "jaxcent-handler")) {
        return DECLINED;
    }

    if ( ! inited ) {
        init_java_connection();
    }
    // Set headers.
    apr_table_setn( r->headers_out, "Cache-Control", "no-cache" );
    apr_table_setn( r->headers_out, "Pragma", "no-cache" );
    apr_table_setn( r->headers_out, "Expires", "0" );

    if (r->header_only) {
        return OK;
    }

    if ( error ) {
        ap_set_content_type(r, "text/plain");
        ap_rputs( error, r );
        return OK;
    }

    ServerList * s = getServer( r->server );
    if ( s->error ) {
        ap_set_content_type(r, "text/plain");
        ap_rputs( s->error, r );
        return OK;
    }

    JNIEnv * jenv;

    if ( vm->AttachCurrentThread( (void **) &jenv, NULL ) != 0 ) {
        ap_set_content_type(r, "text/plain");
        ap_rputs( "jaxcent&21&Could not attach to Java thread", r );
        return OK;
    }

    jobject connector = getConnector( jenv, r->server );
    if ( connector == NULL ) {
        if ( s->error ) {
            ap_set_content_type(r, "text/plain");
            ap_rputs( s->error, r );
        }
        return OK;
    }

    // Read posted multipart data if any
    if ( uploadMethod != NULL ) {
        const char * contentType = apr_table_get(r->headers_in, "Content-Type");
        if ( contentType && strncasecmp( contentType, "multipart/form-data", 19 ) == 0 ) {
            // Process as upload data.
            return processUpload( jenv, r, contentType, connector );

        }
    }
    char postBuf[ 400 ]; // Avoid alloc's for typical Jaxcent messages.
    unsigned int postSize = sizeof( postBuf );
    char * postData = postBuf;
    unsigned int postLen = 0;

    apr_bucket_brigade * brigade = apr_brigade_create( r->pool, r->connection->bucket_alloc );

    bool send = true;
    bool seen_eos = false;

    do {
        apr_bucket *bucket;

        if ( ap_get_brigade(r->input_filters, brigade,
                                AP_MODE_READBYTES, APR_BLOCK_READ,
                                8192) != APR_SUCCESS )
                                break;

        for (bucket = APR_BRIGADE_FIRST(brigade);
             bucket != APR_BRIGADE_SENTINEL(brigade);
             bucket = APR_BUCKET_NEXT(bucket))
        {
            const char *data;
            apr_size_t len;

            if (APR_BUCKET_IS_EOS(bucket)) {
                seen_eos = true;
                break;
            }

            if (APR_BUCKET_IS_METADATA(bucket)) {
                continue;
            }

            if (  apr_bucket_read(bucket, &data, &len, APR_BLOCK_READ) != APR_SUCCESS )
                break;

            if ( len > 0 && data != 0 ) {
                if ( postLen + len >= postSize ) {
                    do {
                        postSize += postSize;
                    } while ( postLen + len >= postSize );
                    if ( postData == postBuf )
                        postData = (char *) malloc( postSize );
                    else
                        postData = (char *) realloc( postData, postSize );
                }
                memcpy( &postData[ postLen ], data, len );
                postLen += len;
            }
        }

        apr_brigade_cleanup(brigade);
    } while (!seen_eos);

    apr_brigade_destroy(brigade);
    postData[ postLen ] = 0;

    jstring jpostData = NULL;
    if ( postLen > 0 ) {
        jpostData = jenv->NewStringUTF( postData );
    }
    if ( postData != postBuf )  // If alloc'ed, free the data buffer
        free( postData );
    if ( jpostData == NULL ) {
        if ( r->args && *r->args ) {
            jpostData = jenv->NewStringUTF( r->args );
        }
    }

    jobjectArray request = jenv->NewObjectArray( 3, jobjectClass, 0 );
    jstring jcookie = NULL;
    const char * cookie_header = apr_table_get(r->headers_in, "Cookie");
    const char * accept_language = apr_table_get(r->headers_in, "Accept-Language");
    const char * remoteAddr = inet_ntoa( r->connection->remote_addr->sa.sin.sin_addr );
    jstring jRemoteAddr = jenv->NewStringUTF( remoteAddr );
    jstring jAcceptLang = NULL;
    if ( accept_language )
        jAcceptLang = jenv->NewStringUTF( accept_language );
    if ( cookie_header != NULL ) {
        // Locate cookie.
        char * cookielist = new char[ strlen( cookie_header ) + 1 ];
        strcpy( cookielist, cookie_header );
        // Look for Jaxcent ID
        char * cookie = strtok( cookielist, ";" );
        while ( cookie != NULL ) {
            while ( *cookie && isspace( *cookie ))
                cookie++;
            if ( strncasecmp( cookie, jaxcentCookieId, sizeof( jaxcentCookieId ) - 1 ) == 0 &&
                cookie[ sizeof( jaxcentCookieId ) - 1 ] == '=' ) {
                // Have jaxcent session cookie
                jcookie = jenv->NewStringUTF( &cookie[ sizeof( jaxcentCookieId ) ] );
                break;
            }
            cookie = strtok( NULL, ";" );
        }
        delete[] cookielist;
    }
    jenv->SetObjectArrayElement( request, 0, jpostData );
    jenv->SetObjectArrayElement( request, 1, jcookie );
    jenv->SetObjectArrayElement( request, 2, 0 );

    jenv->CallVoidMethod( connector, processMethod, request, jRemoteAddr, jAcceptLang );

    jstring jContentType = (jstring) jenv->GetObjectArrayElement( request, 2 );
    if ( jContentType != NULL ) {
        const char * ctype = jenv->GetStringUTFChars( jContentType, 0 );
        ap_set_content_type(r, ctype );
        jenv->ReleaseStringUTFChars( jContentType, ctype );
        jenv->DeleteLocalRef( jContentType );
    } else
        ap_set_content_type(r, "text/plain");

    if ( jpostData )
        jenv->DeleteLocalRef( jpostData );
    if ( jcookie )
        jenv->DeleteLocalRef( jcookie );
    if ( jAcceptLang )
        jenv->DeleteLocalRef( jAcceptLang );

    // Retrieve set-cookie and returned data, send to client.

    jbyteArray resp = (jbyteArray) jenv->GetObjectArrayElement( request, 0 );
    jcookie = (jstring) jenv->GetObjectArrayElement( request, 1 );

    char * cookieValue = NULL;
    if ( jcookie ) {
        const char * cookie = jenv->GetStringUTFChars( jcookie, 0 );
        cookieValue = new char[ strlen( cookie ) + sizeof( jaxcentCookieId ) + 40 ];
        strcpy( cookieValue, jaxcentCookieId );
        strcat( cookieValue, "=" );
        strcat( cookieValue, cookie );
        strcat( cookieValue, ";PATH=/" );
        apr_table_setn( r->headers_out, "Set-Cookie", cookieValue );
    }

    if ( resp != NULL ) {
        int len = jenv->GetArrayLength( resp );
        jbyte * postResult = jenv->GetByteArrayElements( resp, 0 );
        ap_rwrite( postResult, len, r );
        jenv->ReleaseByteArrayElements( resp, postResult, 0 );
    }

    if ( jcookie )
        jenv->DeleteLocalRef( jcookie );
    jenv->DeleteLocalRef( request );
    jenv->DeleteLocalRef( jRemoteAddr );

    if ( cookieValue != NULL )
        delete[] cookieValue;

    if ( resp == NULL ) {
        return HTTP_NOT_FOUND;
    } else
        jenv->DeleteLocalRef( resp );

    return OK;
}

// The parameter handlers.

static const char *set_jvm(cmd_parms *cmd, void *mconfig, const char *arg)
{
    add_parameter( NULL, "JVM", arg );
    return NULL;
}

static const char * set_classpath(cmd_parms *cmd, void *mconfig, const char *arg)
{
    add_parameter( NULL, "Classpath", arg );
    return NULL;
}

static const char * set_config_file(cmd_parms *cmd, void *mconfig, const char *arg)
{
    add_parameter( cmd->server, "JaxcentConfigXML", arg );
    return NULL;
}

static const char * set_log_file(cmd_parms *cmd, void *mconfig, const char *arg)
{
    add_parameter( NULL, "LogFile", arg );
    return NULL;
}

static const char * set_reloadable_classes(cmd_parms *cmd, void *mconfig, const char *arg)
{
    add_parameter( cmd->server, "ReloadableClasses", arg );
    return NULL;
}

static const char * set_display_servlet_message(cmd_parms *cmd, void *dummy, const char *arg)
{
    add_parameter( cmd->server, "DisplayServletMessage", arg );
    return NULL;
}

static const char * set_refresh_url_map(cmd_parms *cmd, void *dummy, const char *arg)
{
    add_parameter( cmd->server, "RefreshURLMap", arg );
    return NULL;
}

static const char * set_keep_alive_timeout(cmd_parms *cmd, void *dummy, const char *arg)
{
    add_parameter( cmd->server, "KeepAliveTimeout", arg );
    return NULL;
}

static const char * set_parameter(cmd_parms *cmd, void *mconfig, const char *arg)
{
    char * value = strchr( arg, '=' );
    if ( value == NULL )
        return NULL;
    int keylen = value - arg;
    if ( keylen <= 0 )
        return NULL;
    value++;
    while ( *value && isspace( *value ))
        value++;
    if ( *value == '\0' )
        return NULL;
    char key[ 400 ];
    if ( keylen >= sizeof( key ))
        keylen = sizeof( key ) - 1;
    memcpy( key, arg, keylen );
    key[ keylen ] = 0;
    add_parameter( cmd->server, key, value );
    return NULL;
}

static void register_hooks(apr_pool_t *p)
{
    ap_hook_handler(jaxcent_handler, NULL, NULL, APR_HOOK_MIDDLE);
}


// Module termination -- but there aren't any clean hooks for it so it isn't called right now.
// apr_pool_cleanup_register, registered during child_init, is supposed to handle this
// situations, but seems to fail to call this if a Java thread is running.

#if 0
static void terminate_module()
{
    JNIEnv * jenv;
    if ( vm != NULL ) {
        if ( connector != NULL && vm->AttachCurrentThread( (void **) &jenv, NULL ) == 0 ) {
            jenv->CallVoidMethod( connector, terminateMethod );
        }
        vm = NULL;
    }
}
#endif
