package sample.yuitree;

import jaxcent.*;
import java.util.*;
import java.text.*;

/**
  * Jaxcent YUI (Yahoo User Interface) tree sample.
  *
  * Shows simple programmatic operations.
  */

public class YuiTreeSample extends YuiTree {

    // Reference to the HTML on the page.

    HtmlPara msgs = new HtmlPara( this, "msgs" );

    HtmlSelect nodesel = new HtmlSelect( this, "nodesel" ) {
        protected void onChange( String value ) {
            nodeSelectionChanged( value );
        }
    };

    HtmlInputButton insertButton = new HtmlInputButton( this, "insert" ) {
        protected void onClick( Map data ) {
            onInsert( data );
        }
    };

    HtmlInputButton expandButton = new HtmlInputButton( this, "expand" ) {
        protected void onClick( Map data ) {
            onExpand( data );
        }
    };

    HtmlInputButton collapseButton = new HtmlInputButton( this, "collapse" ) {
        protected void onClick( Map data ) {
            onCollapse( data );
        }
    };

    int nodeIndex;

    // The constructor builds an initial tree.

    public YuiTreeSample() throws Jaxception
    {
        // The tree must be initialized prior to use.
        // The argument to "init" is the ID of the DIV
        // where the tree is to be created.

        init( "yuitree" );

        // Add some text and date nodes.

        // First argument is a name for the node (to refer in the future.)  Optional.
        // Second argument is the text or date for the node.
        // Third argument is true for editable names.
        // Fourth argument is the parent node, or null to insert at root.

        addTextNode( "node1", "Non Editable", false, null );
        addTextNode( "node2", "<FONT color=\"red\">Non-editable</FONT> Text", false, "node1" );
        addDateNode( "node3", new java.util.Date(), false, "node1" );
        addTextNode( "node4", "Editable", true, null );
        addTextNode( "node5", "Editable <b>Text</b>", true, "node4" );
        addDateNode( "node6", new java.util.Date(), true, "node4" );

        // The "draw" method MUST be called to see the changes on the page!

        draw();

        nodesel.insertOption( -1, "node1" );
        nodesel.insertOption( -1, "node2" );
        nodesel.insertOption( -1, "node3" );
        nodesel.insertOption( -1, "node4" );
        nodesel.insertOption( -1, "node5" );
        nodesel.insertOption( -1, "node6" );

        nodeIndex = 7;

    }

    // Click handler.
    protected void onClick( String nodeName )
    {
        try {
            msgs.setInnerHTML( "<I>Click on node \"" + nodeName + "\"</I>" );
        } catch (Jaxception ex) {
            ex.printStackTrace();
        }
    }

    // Handles selection of the "node to operate upon".

    void nodeSelectionChanged( String node )
    {
        try {
            if ( node.equals( "<Root>" )) {
                msgs.setInnerHTML( "<B>Tree root selected for operations</B>" );
            } else
                msgs.setInnerHTML( "<B>Selected node for operations is \"" + node + "\": " + getNodeLabel( node ) + "</B>" );
        } catch (Jaxception ex) {
            ex.printStackTrace();
        }
    }

    // From a map, retrieves the selected node.

    String getNodeName( Map data )
    {
        String nodesel = (String) data.get( "nodesel" );
        if ( nodesel.startsWith( "0:" ))
            return null;  // Root
        int index = nodesel.indexOf( ':' );
        return nodesel.substring( index+1 );
    }

    // Generates a new nodename, e.g. "node8", "node9" etc.
    String newNodeName()
    {
        String name = "node" + nodeIndex++;
        try {
            nodesel.insertOption( -1, name );
        } catch (Jaxception ex) {
            ex.printStackTrace();
        }
        return name;
    }

    // Handles clicks on the Insert button.

    void onInsert( Map data )
    {
        boolean textNode = ((String) data.get( "type" )).startsWith( "0" );
        String text = (String) data.get( "text" );
        String node = getNodeName( data );
        boolean editable = "on".equals( data.get( "editable" ));
        if ( textNode ) {
           if ( text == null || text.trim().equals( "" )) {
               showMessageDialog( "Please enter text for new node" );
           } else try {
               addTextNode( newNodeName(), text, editable, node );
               draw();
               if ( node == null ) node = "&lt;root&gt;";
               msgs.setInnerHTML( "<B>Inserted new text node under \"" + node + "\"</B>" );
           } catch (Jaxception ex) {
               ex.printStackTrace();
           }
        } else {
           // Check if date is valid.
           Date d = null;
           if ( text == null || text.trim().equals( "" ))
               d = new java.util.Date();
           else try {
               d = DateFormat.getDateInstance( DateFormat.SHORT ).parse( text );
           } catch (Exception ex) {
           }
           if ( d == null ) {
               showMessageDialog( "Invalid Date: \"" + text + "\"" );
           } else try {
               addDateNode( newNodeName(), d, editable, node );
               draw();
               if ( node == null ) node = "&lt;root&gt;";
               msgs.setInnerHTML( "<B>Inserted new date node under \"" + node + "\"</B>" );
           } catch (Jaxception ex) {
               ex.printStackTrace();
           }
        }
    }

    // Handles clicks on the Expand button.

    void onExpand( Map data )
    {
        String node = getNodeName( data );
        try {
            if ( node == null ) {
                expandAll();
                msgs.setInnerHTML( "<B>Expanded tree root</B>" );
            } else {
                expand( node );
                msgs.setInnerHTML( "<B>Expanded node " + node + "</B>" );
            }
        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }

    // Handles clicks on the Collapse button.

    void onCollapse( Map data )
    {
        try {
            String node = getNodeName( data );
            if ( node == null ) {
                collapseAll();
                msgs.setInnerHTML( "<B>Collapsed tree root</B>" );
            } else {
                collapse( node );
                msgs.setInnerHTML( "<B>Collapsed node " + node + "</B>" );
            }
        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }
}
