/*
 * Copyright (c) 2008, Desiderata Software
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 * Redistributions of source code must retain the above copyright notice, this list
 * of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright notice, this
 * list of conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution. 
 * 
 * Neither the name of Desiderata Software nor the names of its contributors may be
 * used to endorse or promote products derived from this software without specific
 * prior written permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *
 * Module Name:           JaxcentConnector.cpp
 * Original Author:       Mukesh Prasad
 * Brief Description:     IIS ISAPI Filter for Jaxcent.
 * Change History:
 *
 *     1/23/2008  MP      Initial version.
 *
 */

#include "stdafx.h"

const char defaultHeaders[] = "Content-Type: text/plain\r\nCache-Control: no-cache\r\nPragma: no-cache\r\nExpires: 0\r\n\r\n";
const char jaxcentCookieId[] = "JAXCENT_ISAPI_SESSIONID";

HINSTANCE g_hInstance;
bool inited = false;
long lastFileTime;
char * error = NULL;
struct ParamList {
	char * key;
	char * value;
	ParamList * link;
};

ParamList * params = NULL;
ParamList * lastParam = NULL;
typedef jint (JNICALL * CreateJVM)(JavaVM **, void **, JavaVMInitArgs *);
JavaVM * vm = 0;
jobject connector;
jmethodID processMethod;
jmethodID terminateMethod;
jmethodID uploadInitMethod;
jmethodID uploadMethod;
jmethodID uploadEndMethod;
jclass jobjectClass;

BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
	if ( ul_reason_for_call == DLL_PROCESS_ATTACH )
	{
		g_hInstance = (HINSTANCE) hModule;
	}
    return TRUE;
}

void setError( const char * fmt, ... )
{
	va_list marker;
	char buffer[ 2000 ];

	va_start( marker, fmt );
	int len = vsprintf( buffer, fmt, marker );
	va_end( marker );
	if ( error != NULL )
		delete[] error;
	error = new char[ strlen( buffer ) + 1 ];
	strcpy( error, buffer );
}

void DeleteParams()
{
	while ( params ) {
		ParamList * pl = params;
		params = params->link;
		delete[] pl->key;
		delete[] pl->value;
		delete pl;
	}
}

void InitConnector()
{
	char thisDllPath[ _MAX_PATH ];
	char inifile[ _MAX_PATH ];
	char * ptr, * key, * value;
	char * jvm, * classpath, * configFile;
	char * logfile;

	if ( inited && error == NULL )
		return;
	// Read the .ini file.
	::GetModuleFileName( g_hInstance, thisDllPath, sizeof thisDllPath );
	char * dllPath = thisDllPath;
	if ( strncmp( dllPath, "\\\\", 2 ) == 0 ) {
		dllPath = strchr( dllPath, ':' );
		dllPath--;
	}
	strcpy( inifile, dllPath );
	strcpy( &inifile[ strlen( inifile ) - 4 ], ".ini" );
	struct _stat stat;
	if ( _stat( inifile, &stat ) != 0 ) {
		setError( "Initialization file \"%s\" not found", inifile );
		return;
	}
	if ( inited ) {
		// Did the .ini file change, or has the VM already been loaded?
		if (( lastFileTime == stat.st_mtime ) || ( vm != 0 )) {
			return;
		}
	}
	FILE * fp = fopen( inifile, "r" );
	if ( fp == NULL ) {
		setError( "Initialization file \"%s\" not found", inifile );
		return;
	}
	if ( error ) {
		delete[] error;
		error = NULL;
		DeleteParams();
	}
	inited = true;
	lastFileTime = stat.st_mode;
	jvm = NULL;
	classpath = NULL;
	configFile = NULL;
	logfile = NULL;
	char buf[ 512 ];
	while ( fgets( buf, sizeof buf, fp )) {
		// Parse line.
		ptr = buf;
		while ( *ptr && isspace( *ptr ))
			ptr++;
		if ( *ptr == '#' || *ptr == ';' )
			continue;
		value = strchr( ptr, ':' );
		if ( value == NULL )
			continue;
		key = ptr;
		ptr = value++;
		while ( ptr > key && isspace( ptr[-1] ))
			ptr--;
		*ptr = '\0';
		while ( *value && isspace( *value ))
			value++;
		ptr = &value[ strlen( value ) ];
		while ( ptr > value && isspace( ptr[-1] ))
			ptr--;
		*ptr = '\0';
		// Save the key/value.
		ParamList * pl = new ParamList;
		pl->key = new char[ strlen( key ) + 1 ];
		pl->value = new char[ strlen( value ) + 1 ];
		strcpy( pl->key, key );
		strcpy( pl->value, value );
		pl->link = NULL;
		if ( lastParam == NULL )
			params = pl;
		else
			lastParam->link = pl;
		lastParam = pl;
		if ( stricmp( key, "jvm" ) == 0 ) {
			// Have JVM
			jvm = pl->value;
		} else if ( stricmp( key, "classpath" ) == 0 ) {
			classpath = pl->value;
		} else if ( stricmp( key, "jaxcentconfigxml" ) == 0 ) {
			configFile = pl->value;
		} else if ( stricmp( key, "logfile" ) == 0 ) {
			logfile = pl->value;
		}
	}
	fclose( fp );
	if ( jvm == NULL ) {
		setError( "JVM is not specified in ini file \"%s\"", inifile );
		return;
	}
	if ( strlen( jvm ) < 5 ||
		stricmp( &jvm[ strlen(jvm)-4 ], ".dll" ) != 0 ) {
		setError( "JVM must be a .DLL file. (Found \"%s\" in file %s)", jvm, inifile );
		return;
	}
	if ( classpath == NULL ) {
		setError( "CLASSPATH is not specified in ini file \"%s\"", inifile );
		return;
	}
	if ( configFile == NULL ) {
		setError( "JaxcentConfigXML is not specified in ini file \"%s\"", inifile );
		return;
	}
	if ( _access( configFile, 04 ) != 0 ) {
		setError( "Cannot read JaxcentConfigXML file \"%s\"", configFile );
		return;
	}
	if ( logfile == NULL ) {
		setError( "LogFile is not specified in ini file \"%s\"", inifile );
		return;
	}
	if ( _access( logfile, 0 ) != 0 ) {
		FILE * fp = fopen( logfile, "wb" );
		if ( fp )
			fclose( fp );
	}
	if ( _access( logfile, 02 ) != 0 ) {
		setError( "Cannot write LogFile \"%s\" (place in same directory as script or under wwwroot)", logfile );
		return;
	}
	// Load the JVM.
	HMODULE jvmLib = LoadLibrary( jvm );
	if ( jvmLib == NULL ) {
		void * lpMsgBuf;
		if ( FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
			NULL, GetLastError(), MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),(LPTSTR) &lpMsgBuf, 0, 0 ) == 0 )
			setError( "Could not load JVM \"%s\": %s", jvm );
		else {
			setError( "Could not load JVM \"%s\": %s", jvm, lpMsgBuf );
			LocalFree( lpMsgBuf );
		}
		return;
	}
	CreateJVM createJVM = (CreateJVM) ::GetProcAddress( jvmLib, "JNI_CreateJavaVM" );
	if ( createJVM == NULL ) {
		setError( "JVM DLL \"%s\" does not contain a valid JVM", jvm );
		return;
	}
	JavaVMInitArgs vm_args;
	vm_args.version = JNI_VERSION_1_2;
	JavaVMOption optClasspath;
	char * classpathOption = new char[ strlen( classpath ) + 22 ];
	strcpy( classpathOption, "-Djava.class.path=" );
    strcat( classpathOption, classpath );
	optClasspath.optionString = classpathOption;
	optClasspath.extraInfo = 0;
	vm_args.nOptions = 1;
	vm_args.options = &optClasspath;
	JNIEnv * jenv;
	int result = createJVM( &vm, (void **) &jenv, &vm_args );
	delete[] classpathOption;
	if ( result < 0 ) {
		setError( "Java VM Creation from \"%s\" failed", jvm );
		return;
	}
	// Load the object
	if ( jenv->FindClass( "jaxcentFramework/JaxcentFramework" ) == NULL ) {
		setError( "Could not find class \"jaxcentFramework.JaxcentFramework\" in classpath\r\n     \"%s\"\r\nMake sure to include path to jaxcentFramework JAR file.\r\n\r\nExtension must be unloaded and reloaded to read the new classpath.", classpath );
		return;
	}
	jclass connectorClass = jenv->FindClass( "jaxcentConnector/JaxcentHttpConnector" );
	if ( connectorClass == NULL ) {
		setError( "Could not find class \"jaxcentConnector.JaxcentHttpConnector\" in classpath\r\n     \"%s\"\r\nMake sure to include path to jaxcentConnector JAR file.\r\n\r\nExtension must be unloaded and reloaded to read the new classpath.", classpath );
		return;
	}
	// Get the ctor.
	jmethodID ctor = jenv->GetMethodID( connectorClass, "<init>", "([Ljava/lang/String;[Ljava/lang/String;)V" );
	if ( ctor == NULL ) {
		setError( "Internal error in class jaxcentConnector.JaxcentHttpConnector (ctor not found)" );
		return;
	}
	jmethodID initMethod = jenv->GetMethodID( connectorClass, "init", "()V" );
	if ( initMethod == NULL ) {
		setError( "Internal error in class jaxcentConnector.JaxcentHttpConnector (\"init\" not found)" );
		return;
	}
	processMethod = jenv->GetMethodID( connectorClass, "process", "([Ljava/lang/Object;Ljava/lang/String;Ljava/lang/String;)V" );
	if ( processMethod == NULL ) {
		setError( "Internal error in class jaxcentConnector.JaxcentHttpConnector (\"init\" not found)" );
		return;
	}
	terminateMethod = jenv->GetMethodID( connectorClass, "terminate", "()V" );
	if ( terminateMethod == NULL ) {
		setError( "Internal error in class jaxcentConnector.JaxcentHttpConnector (\"terminate\" not found)" );
		return;
	}
	uploadInitMethod = jenv->GetMethodID( connectorClass, "onUploadInit", "(Ljava/lang/String;Ljava/lang/String;I)Ljava/lang/Object;" );
	if ( jenv->ExceptionOccurred()) {
		uploadInitMethod = 0;
		jenv->ExceptionClear();
	}
	uploadEndMethod = jenv->GetMethodID( connectorClass, "onUploadEnd", "(Ljava/lang/Object;)[B" );
	if ( jenv->ExceptionOccurred()) {
		uploadEndMethod = 0;
		jenv->ExceptionClear();
	}
	uploadMethod = jenv->GetMethodID( connectorClass, "onUpload", "(Ljava/lang/Object;[B)Z" );
	if ( jenv->ExceptionOccurred()) {
		uploadMethod = 0;
		jenv->ExceptionClear();
	}
	jmethodID logFileMethod = jenv->GetStaticMethodID( connectorClass, "setLogfile", "(Ljava/lang/String;)Ljava/lang/String;" );
	if ( logFileMethod == NULL ) {
		setError( "Internal error in class jaxcentConnector.JaxcentHttpConnector (\"setLogfile\" not found)" );
		return;
	}
	jstring jlogfile = jenv->NewStringUTF( logfile );
	jstring jlogresult = (jstring) jenv->CallStaticObjectMethod( connectorClass, logFileMethod, jlogfile );
	if ( jlogresult != NULL ) {
		const char * logresult = jenv->GetStringUTFChars( jlogresult, NULL );
		setError( logresult );
		jenv->ReleaseStringUTFChars( jlogresult, logresult );
		jenv->DeleteLocalRef( jlogresult );
	}
	jenv->DeleteLocalRef( jlogfile );

	int nCount = 0;
	ParamList * pl;
	for ( pl = params; pl; pl = pl->link )
		nCount++;
	jclass stringClass = jenv->FindClass( "java/lang/String" );
	jobjectArray keys = jenv->NewObjectArray( nCount, stringClass, NULL );
	jobjectArray values = jenv->NewObjectArray( nCount, stringClass, NULL );
	int index = 0;
	for ( pl = params; pl; pl = pl->link ) {
		jenv->SetObjectArrayElement( keys, index, jenv->NewStringUTF( pl->key ));
		jenv->SetObjectArrayElement( values, index, jenv->NewStringUTF( pl->value ));
		index++;
	}
	connector = jenv->NewObject( connectorClass, ctor, keys, values );
	if ( connector == NULL ) {
		setError( "Internal error: Could not instantiate jaxcentConnector.JaxcentHttpConnector" );
		return;
	}
	connector = jenv->NewGlobalRef( connector );
	jenv->CallVoidMethod( connector, initMethod );
	jobjectClass = (jclass) jenv->NewGlobalRef( jenv->FindClass( "java/lang/Object" ));
}

BOOL WINAPI GetExtensionVersion( HSE_VERSION_INFO* pVer )
{
	InitConnector();
	DeleteParams();

    pVer->dwExtensionVersion = MAKELONG(HSE_VERSION_MINOR, HSE_VERSION_MAJOR);
    strcpy(pVer->lpszExtensionDesc, "Jaxcent Connector Server Extension, Copyright (C) Desiderata Software, 2008");
 
	return (TRUE);
}

static bool onUpload( JNIEnv * jenv, char * data, int len, jobject ref )
{
	bool ret = true;
	if ( len > 0 && data != 0 ) {
		jbyteArray jdata = jenv->NewByteArray( len );
		jbyte * jbuf = jenv->GetByteArrayElements( jdata, NULL );
		memcpy( jbuf, data, len );
		jenv->ReleaseByteArrayElements( jdata, jbuf, 0 );
		if ( jenv->CallBooleanMethod( connector, uploadMethod, ref, jdata ) == JNI_FALSE )
			ret = false;
		jenv->DeleteLocalRef( jdata );
	}
	return ret;
}

static void processUpload( JNIEnv * jenv, LPEXTENSION_CONTROL_BLOCK pECB )
{
	jstring jargs = 0;
	if ( pECB->lpszQueryString && * pECB->lpszQueryString ) {
		jargs = jenv->NewStringUTF( pECB->lpszQueryString );
	}
	int clen = pECB->cbTotalBytes;

	jstring jcontentType = jenv->NewStringUTF( pECB->lpszContentType );
	jobject ref = jenv->CallObjectMethod( connector, uploadInitMethod, jargs, jcontentType, clen );
	jenv->DeleteLocalRef( jcontentType );

	char buffer[ 8192 ];
	if ( clen > 0 ) {
		bool send = true;
		if ( pECB->cbAvailable > 0 ) {
			send = onUpload( jenv, (char *) pECB->lpbData, pECB->cbAvailable, ref );
			clen -= pECB->cbAvailable;
		}
		while ( clen > 0 ) {
			DWORD n = sizeof buffer;
			if ( ! pECB->ReadClient( pECB->ConnID, buffer, &n )) {
				break;
			}
			if ( send )
				send = onUpload( jenv, buffer, n, ref );
			clen -= n;
		}
	}


	jbyteArray resp = NULL;
	resp = (jbyteArray) jenv->CallObjectMethod( connector, uploadEndMethod, ref );
	if ( ref )
		jenv->DeleteLocalRef( ref );

	char * headers = (char *) defaultHeaders;
	DWORD len = strlen( headers );
	pECB->ServerSupportFunction(
		pECB->ConnID,
		HSE_REQ_SEND_RESPONSE_HEADER,
		NULL,
		&len,
		(DWORD *) headers );
 
	if ( resp != NULL ) {

		int len = jenv->GetArrayLength( resp );
		if ( len > 0 ) {
			jbyte * postResult = jenv->GetByteArrayElements( resp, 0 );
			pECB->WriteClient( pECB->ConnID, (void *) postResult, (DWORD *) &len, HSE_IO_SYNC );
			jenv->ReleaseByteArrayElements( resp, postResult, 0 );
		}
		jenv->DeleteLocalRef( resp );
	}

	if ( jargs )
		jenv->DeleteLocalRef( jargs );
}

DWORD WINAPI HttpExtensionProc( LPEXTENSION_CONTROL_BLOCK pECB )
{
	if ( error ) {
		InitConnector();
		DeleteParams();
	}

	DWORD hseStatus = HSE_STATUS_SUCCESS;

	if ( error ) {
		pECB->ServerSupportFunction(
			pECB->ConnID,
			HSE_REQ_SEND_RESPONSE_HEADER,
			NULL,
			NULL,
			(DWORD *) defaultHeaders );

		DWORD len = strlen( error );
		pECB->WriteClient( pECB->ConnID, error, (DWORD *) &len, HSE_IO_SYNC );
		return hseStatus;
	}

	JNIEnv * jenv;

	if ( vm->AttachCurrentThread( (void **) &jenv, NULL ) != 0 ) {
		pECB->ServerSupportFunction(
			pECB->ConnID,
			HSE_REQ_SEND_RESPONSE_HEADER,
			NULL,
			NULL,
			(DWORD *) defaultHeaders );

		char * err = "jaxcent&21&Could not attach to Java thread";
		DWORD len = strlen( err );
		pECB->WriteClient( pECB->ConnID, err, (DWORD *) &len, HSE_IO_SYNC );
		return hseStatus;
	}

	// Read posted data
	if ( uploadMethod != NULL ) {
		if ( pECB->lpszContentType && strnicmp( pECB->lpszContentType, "multipart/form-data", 19 ) == 0 ) {
			// Process as upload data.
			processUpload( jenv, pECB );
			return hseStatus;
		}
	}

	char remoteAddr[ 32 ];
	char buffer[ 4096 ];

	DWORD n = sizeof( remoteAddr );
	if ( ! pECB->GetServerVariable( pECB->ConnID, "REMOTE_ADDR", remoteAddr, &n ))
		*remoteAddr = '\0';
	n = sizeof( buffer );
	jstring jcookie = 0;
	jstring jAcceptLang = 0;
	char * cookie = NULL;
	if ( pECB->GetServerVariable( pECB->ConnID, "ALL_HTTP", buffer, &n )) {
		char * toks = strtok( buffer, "\r\n" );
		while ( toks != NULL ) {
			if ( strnicmp( toks, "HTTP_ACCEPT_LANGUAGE:", 21 ) == 0 ) {
				char * al = &toks[21];
				while ( *al && isspace( *al ))
					al++;
				jAcceptLang = jenv->NewStringUTF( al );
			} else if ( strnicmp( toks, "HTTP_COOKIE:", 12 ) == 0 ) {
				cookie = &toks[12];
			}
			toks = strtok( NULL, "\r\n" );
		}
	}
	// Did we find a cookielist?
	if ( cookie ) {
		// Look for Jaxcent ID
		while ( cookie && *cookie ) {
			while ( *cookie && isspace( *cookie ))
					cookie++;
			char * term = strchr( cookie, ';' );
			if ( term == NULL )
				term = &cookie[ strlen( cookie ) ];
			if ( strnicmp( cookie, jaxcentCookieId, sizeof( jaxcentCookieId ) - 1 ) == 0 &&
				cookie[ sizeof( jaxcentCookieId ) - 1 ] == '=' ) {
				// Have jaxcent session cookie
				*term = 0;
				jcookie = jenv->NewStringUTF( &cookie[ sizeof( jaxcentCookieId ) ] );
				break;
			}
			if ( *term == '\0' )
				break;
			cookie = term+1;
		}
	}

	char * respcode = NULL;

	int clen = pECB->cbTotalBytes;
	jstring jpostData = NULL;
	if ( clen > 0 ) {
		char * postData = new char[ clen+1 ];
		char * ptr = postData;
		if ( pECB->cbAvailable > 0 ) {
			int len = pECB->cbAvailable;
			if ( len > clen )
				len = clen;
			memcpy( ptr, (char *) pECB->lpbData, len );
			ptr += n;
			clen -= n;
		}
		while ( clen > 0 ) {
			n = sizeof buffer;
			if ( ! pECB->ReadClient( pECB->ConnID, buffer, &n )) {
				break;
			}
			memcpy( ptr, buffer, n );
			ptr += n;
			clen -= n;
		}
		postData[pECB->cbTotalBytes] = 0;
		jpostData = jenv->NewStringUTF( postData );
	}
	if ( jpostData == NULL ) {
		if ( pECB->lpszQueryString && * pECB->lpszQueryString )
			jpostData = jenv->NewStringUTF( pECB->lpszQueryString );
	}

	jobjectArray request = jenv->NewObjectArray( 3, jobjectClass, 0 );
	jstring jRemoteAddr = jenv->NewStringUTF( remoteAddr );
	jenv->SetObjectArrayElement( request, 0, jpostData );
	jenv->SetObjectArrayElement( request, 1, jcookie );
	jenv->SetObjectArrayElement( request, 2, 0 );

	jenv->CallVoidMethod( connector, processMethod, request, jRemoteAddr, jAcceptLang );

	if ( jpostData )
		jenv->DeleteLocalRef( jpostData );
	if ( jcookie )
		jenv->DeleteLocalRef( jcookie );
	jenv->DeleteLocalRef( jRemoteAddr );
	if ( jAcceptLang )
		jenv->DeleteLocalRef( jAcceptLang );

	jbyteArray resp = (jbyteArray) jenv->GetObjectArrayElement( request, 0 );
	jcookie = (jstring) jenv->GetObjectArrayElement( request, 1 );
	jstring jContentType = (jstring) jenv->GetObjectArrayElement( request, 2 );

	// If text response is null, set respcode to "404 not found"
	if ( resp == NULL ) {
		respcode = "404 Not Found";
	}

	char * headers = (char *) defaultHeaders;

	if ( jContentType ) {
		const char * ctype = jenv->GetStringUTFChars( jcookie, 0 );
		headers = new char[ strlen( ctype ) + sizeof( defaultHeaders ) ];
		strcpy( headers, "Content-Type: " );
		strcat( headers, ctype );
		strcat( headers, "\r\nCache-Control: no-cache\r\nPragma: no-cache\r\nExpires: 0\r\n\r\n" );
		jenv->ReleaseStringUTFChars( jContentType, ctype );
		jenv->DeleteLocalRef( jContentType );
	}
	// If there is a cookie id, add it to the hbuf.
	if ( jcookie ) {
		const char * cookie = jenv->GetStringUTFChars( jcookie, 0 );
		headers = new char[ strlen( cookie ) + sizeof( defaultHeaders ) + sizeof( jaxcentCookieId ) + 40 ];
		strcpy( headers, "Set-Cookie: " );
		strcat( headers, jaxcentCookieId );
		strcat( headers, "=" );
		strcat( headers, cookie );
		strcat( headers, ";PATH=/\r\n" );
		strcat( headers, defaultHeaders );
		jenv->ReleaseStringUTFChars( jcookie, cookie );
		jenv->DeleteLocalRef( jcookie );
	}

	DWORD len = strlen( headers );
	pECB->ServerSupportFunction(
		pECB->ConnID,
		HSE_REQ_SEND_RESPONSE_HEADER,
		(void *) respcode,
		&len,
		(DWORD *) headers );
 
	if ( resp != 0 ) {
		len = jenv->GetArrayLength( resp );
		jbyte * data = jenv->GetByteArrayElements( resp, 0 );
		pECB->WriteClient( pECB->ConnID, (void *) resp, (DWORD *) &len, HSE_IO_SYNC );
		jenv->ReleaseByteArrayElements( resp, data, 0 );
		jenv->DeleteLocalRef( resp );
	}

	jenv->DeleteLocalRef( request );

	if ( headers != defaultHeaders )
		delete[] headers;

	return hseStatus;
}

BOOL WINAPI TerminateExtension( DWORD dwFlags )
{
	JNIEnv * jenv;
	if ( vm != NULL ) {
		if ( connector != NULL && vm->AttachCurrentThread( (void **) &jenv, NULL ) == 0 ) {
			jenv->CallVoidMethod( connector, terminateMethod );
		}
		// vm->DestroyJavaVM(); 
	}
	return TRUE;
}
